from typing import Optional
from base64 import b64encode
from aiohttp import ClientSession
from datetime import datetime, timezone

from .base_auth_flow import BaseAuthFlow
from .token import Token


class ClientCredentialsFlow(BaseAuthFlow):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        token: Optional[Token] = None
    ) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.token = token

    async def get_access_token(self, http: ClientSession) -> Token:
        API_URL = "https://accounts.spotify.com/api/token"
        GRANT_TYPE = {"grant_type": "client_credentials"}

        if self.token and self.token.expires_at > datetime.now(timezone.utc):
            return self.token

        async with http.post(
            API_URL,
            data=GRANT_TYPE,
            headers=await self.make_auth_header()
        ) as r:
            return await Token.from_dict(await r.json())

    async def make_auth_header(self) -> dict:
        return {
            "Authorization": "Basic %s" % b64encode(
                f"{self.client_id}:{self.client_secret}".encode()
            ).decode()
        }



class SpotifyException(Exception):
    """
    General Spotify exception.
    Packs together all the possible errors Spotify could give
    """

    def __init__(self, status: int, message: str) -> None:
        self.status = status
        self.message = message

    def __repr__(self) -> str:
        return "<SpotifyException(status={0}, message=\"{1}\")>".format(
            self.status,
            self.message
        )



from __future__ import annotations
from typing import List, Optional
from datetime import datetime, timedelta, timezone


class Token:
    def __init__(
        self,
        access_token: str,
        token_type: str,
        scopes: Optional[List[str]],
        expires_in: int,
        expires_at: datetime,
        refresh_token: Optional[str] = None
    ) -> None:
        self.access_token = access_token
        self.token_type = token_type
        self.scopes = scopes
        self.expires_in = expires_in
        self.expires_at = expires_at
        self.refresh_token = refresh_token

    @classmethod
    async def from_dict(cls, d: dict) -> Token:
        expires_at = datetime.now(
            timezone.utc
        ) + timedelta(seconds=d["expires_in"])

        return cls(
            access_token=d["access_token"],
            token_type=d["token_type"],
            scopes=d.get("scope", "").split() or None,
            expires_in=d["expires_in"],
            expires_at=expires_at,
            refresh_token=d.get("refresh_token")
        )



from __future__ import annotations
from typing import List

import spotipy2
from spotipy2.types import Artist, Track


class ArtistMethods:
    async def get_artists(
        self: spotipy2.Spotify,
        artist_ids: List[str]
    ) -> List[Artist]:
        artists = await self._get(
            "artists",
            params={"ids": ",".join([self.get_id(i) for i in artist_ids])}
        )
        return [Artist.from_dict(track) for track in artists["artists"]]

    async def get_artist(self: spotipy2.Spotify, artist_id: str) -> Artist:
        return Artist.from_dict(
            await self._get(f"artists/{self.get_id(artist_id)}")
        )

    async def get_artist_top_tracks(
        self: spotipy2.Spotify, artist_id: str, market: str = "US"
    ) -> List[Track]:
        top_tracks = await self._get(
            f"artists/{self.get_id(artist_id)}/top-tracks",
            params={"country": market}
        )
        return [Track.from_dict(track) for track in top_tracks["tracks"]]



from __future__ import annotations


class SimplifiedArtist:
    def __init__(
        self,
        external_urls: dict,
        href: str,
        id: str,
        name: str,
        type: str,
        uri: str,
        **kwargs
    ):
        self.external_urls = external_urls
        self.href = href
        self.id = id
        self.name = name
        self.type = type
        self.uri = uri

    @classmethod
    def from_dict(cls, d: dict) -> SimplifiedArtist:
        return cls(**d)



class BaseAuthFlow:
    def __init__(self) -> None:
        raise TypeError(
            "Base types can only be used for type checking purposes: "
            "you tried to use a base type instance as argument, "
            "but you need to instantiate one of its subclass instead. "
        )



import re

from .artists import ArtistMethods
from .tracks import TrackMethods


class Methods(ArtistMethods, TrackMethods):
    @staticmethod
    def get_id(s: str) -> str:
        if m := re.search("(?!.*/).+", s):
            return m[0].split("?")[0]
        else:
            raise ValueError("ID not valid")



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()

setup(
    name="spotipy2",
    version="0.3",
    description="The next generation Spotify Web API wrapper for Python",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/CyanBook/spotipy2",
    download_url="https://github.com/CyanBook/spotipy2/releases/latest",
    author="CyanBook",
    license="LGPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    keywords="spotify spotipy api wrapper client library oauth",
    project_urls={
        "Tracker": "https://github.com/CyanBook/spotipy2/issues",
        "Community": "https://github.com/CyanBook/spotipy2",
        "Source": "https://t.me/spotipy2"
    },
    python_requires="~=3.7",
    packages=find_packages()
)



from __future__ import annotations
from typing import List

import spotipy2
from spotipy2.types import Track


class TrackMethods:
    async def get_tracks(
        self: spotipy2.Spotify,
        track_ids: List[str]
    ) -> List[Track]:
        tracks = await self._get(
            "tracks",
            params={"ids": ",".join([self.get_id(i) for i in track_ids])}
        )
        return [Track.from_dict(track) for track in tracks["tracks"]]

    async def get_track(self: spotipy2.Spotify, track_id: str) -> Track:
        return Track.from_dict(
            await self._get(f"tracks/{self.get_id(track_id)}")
        )



from __future__ import annotations
from typing import Optional, List

from spotipy2.types import SimplifiedArtist


class Track:
    def __init__(
        self,
        album,
        artists,
        disc_number: int,
        duration_ms: int,
        explicit: bool,
        external_ids,
        external_urls,
        href: str,
        id: str,
        is_local: bool,
        name: str,
        popularity: int,
        preview_url: str,
        track_number: int,
        type: str,
        uri: str,
        available_markets: Optional[List[str]] = None,
        is_playable: bool = None,
        linked_from: bool = None,
        restrictions: dict = None,
        **kwargs
    ):
        self.album = album
        self.artists = [SimplifiedArtist.from_dict(a) for a in artists]
        self.available_markets = available_markets
        self.disc_number = disc_number
        self.duration_ms = duration_ms
        self.explicit = explicit
        self.external_ids = external_ids
        self.external_urls = external_urls
        self.href = href
        self.id = id
        self.is_local = is_local
        self.name = name
        self.popularity = popularity
        self.preview_url = preview_url
        self.track_number = track_number
        self.type = type
        self.uri = uri
        self.is_playable = is_playable
        self.linked_from = linked_from
        self.restrictions = restrictions

    @classmethod
    def from_dict(cls, d: dict) -> Track:
        return cls(**d)



from __future__ import annotations
from typing import Optional
from aiohttp import ClientSession

from spotipy2.auth import ClientCredentialsFlow
from spotipy2.methods import Methods
from spotipy2.exceptions import SpotifyException


class Spotify(Methods):
    API_URL = "https://api.spotify.com/v1/"

    def __init__(
        self,
        auth_flow: ClientCredentialsFlow,
        *args,
        **kwargs
    ) -> None:
        self.auth_flow = auth_flow
        self.http = ClientSession(*args, **kwargs)

    async def _req(
        self,
        method: str,
        endpoint: str,
        params: Optional[dict] = None
    ) -> dict:
        token = await self.auth_flow.get_access_token(self.http)
        headers = {"Authorization": f"Bearer {token.access_token}"}

        async with self.http.request(
            method,
            f"{self.API_URL}{endpoint}",
            params=params,
            headers=headers
        ) as r:
            json = await r.json()

            try:
                assert r.status == 200
            except AssertionError:
                raise SpotifyException(
                    json["error"]["status"],
                    json["error"]["message"]
                )
            else:
                return json

    async def _get(self, endpoint: str, params: Optional[dict] = None) -> dict:
        return await self._req("GET", endpoint, params)

    async def stop(self) -> None:
        await self.http.close()

    async def __aenter__(self) -> Spotify:
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.stop()
